/*-------------------------------------------------------------------*/
/* CBGKM.C        Marko Tuononen                                     */
/*                                                                   */
/* Console interface for the global K-means algorithm                */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "CBGKM"
#define VersionNumber  "Version 0.11"
#define LastUpdated    "08.09.04"
#define MAXFILENAME    64    /* maximum length of filename */
#define FACTFILE       "cbgkm.fac"

/* ------------------------------------------------------------------- */

#include "parametr.c"
#include "cb.h"
#include "file.h"
#include "interfc.h"
#include "memctrl.h"
#include "gkm.h"


/* ======================== PRINT ROUTINES =========================== */

void PrintInfo(void)
{
  PrintMessage("\n%s   %s   %s\n", ProgName, VersionNumber, LastUpdated);
  PrintMessage("Console interface for the global K-means algorithm.\n");
  PrintMessage("Usage: %s [%coption] ", ProgName, OPTION_SYMBOL);
  PrintMessage("<training set> [initial locations] <result codebook>\n");
  PrintMessage("For example: %s bridge tmp\n", ProgName);
  PrintMessage("\n  Options:\n");
  PrintOptions();
  PrintMessage("\n");
}  /* PrintInfo() */

/* ------------------------------------------------------------------ */

void PrintInitialData(int quietLevel, char *TSName, char *CBName, 
char *InitialCBName, int clus, int kmIter)
{
  char str[127];

  if (quietLevel > 1)
  {
    GKMInfo(str);

    PrintMessage("\n%s\n\n", str);
    PrintMessage("Training Set\t\t = %s\n", TSName);
    if (*InitialCBName)
      PrintMessage("Candidate locations\t = %s\n", InitialCBName);
    PrintMessage("Result Codebook\t\t = %s\n", CBName);
    PrintMessage("Codebook size\t\t = %d\n", clus);
    PrintMessage("K-means iterations\t = %d\n", kmIter);
  }
}  /* PrintInitialData() */


/* ===========================  MAIN  ================================ */

int main(int argc, char* argv[])
{
  TRAININGSET   TS;
  CODEBOOK*     CentroidLocations;
  CODEBOOK      InitialCB;
  CODEBOOK      CB;
  PARTITIONING  P;
  char          TSName[MAXFILENAME] = {'\0'};
  char          InitialCBName[MAXFILENAME] = {'\0'};
  char          CBName[MAXFILENAME] = {'\0'};
  ParameterInfo paraminfo[3] = { { TSName,  FormatNameTS, 0, INFILE },
                                 { InitialCBName, FormatNameCB, 1, INFILE  },
                                 { CBName,  FormatNameCB, 0, OUTFILE } };
  
  ParseParameters(argc, argv, 3, paraminfo);

  /* training set doesn't exist */
  if (!ExistFile(TSName)) 
  {
    ErrorMessage("\nERROR: File doesn't exist: %s\n\n", TSName);
    ExitProcessing(FATAL_ERROR);
  }  

  /* result codebook exists and we are told not to overwrite */
  if (ExistFile(CBName) && !Value(OverWrite))
  {
    ErrorMessage("\nERROR: File already exists: %s\n\n", CBName);
    ExitProcessing(FATAL_ERROR);
  }

  ReadTrainingSet(TSName, &TS);

  /* codebook cannot contain more vectors than training set */
  if (BookSize(&TS) < Value(Clusters))  
  {
    ErrorMessage("\nERROR: Number of vectors in training set ");
    ErrorMessage("(%d) < number of clusters ", BookSize(&TS));
    ErrorMessage("(%d)!\n\n", Value(Clusters));
    FreeCodebook(&TS);
    ExitProcessing(FATAL_ERROR);
  }

  if (*InitialCBName)
  {
    if (!ExistFile(InitialCBName))
    {
      ErrorMessage("\nERROR: File doesn't exist: %s\n\n", InitialCBName);
      ExitProcessing(FATAL_ERROR);
    }
    ReadCodebook(InitialCBName, &InitialCB);
    CentroidLocations = &InitialCB;
  }
  else CentroidLocations = &TS;

  CreateNewCodebook(&CB, Value(Clusters), &TS);
  CreateNewPartitioning(&P, &TS, Value(Clusters));

  PrintInitialData(Value(QuietLevel), TSName, CBName, InitialCBName,
    Value(Clusters), Value(KMeansIterations));
    
  if ( PerformGKM(&P, &CB, CentroidLocations, &TS, Value(KMeansIterations), 
       Value(QuietLevel), Value(SaveIntermediateResults)) )
  {
    ErrorMessage("ERROR: Clustering failed!\n");
    FreeCodebook(&TS);
    if (CentroidLocations == &InitialCB) FreeCodebook(&InitialCB);
    FreeCodebook(&CB);
    FreePartitioning(&P);
    ExitProcessing(FATAL_ERROR);
  }

  WriteCodebook(CBName, &CB, Value(OverWrite));

  FreeCodebook(&TS);
  FreeCodebook(&CB);
  FreePartitioning(&P);
  checkmemory();
 
  return 0;
}  /* main() */

