/*-------------------------------------------------------------------*/
/* GKM.C        Marko Tuononen                                       */
/*                                                                   */
/* The global K-means algorithm presented in                         */
/*                                                                   */
/*   "The global k-means clustering algorithm"                       */
/*   Pattern Recognition, 36, 451-461, 2003.                         */
/*   Aristidis Likas, Nikos Vlassis and Jakob J. Verbeek.            */
/*                                                                   */
/* K-means -operation uses activity detection presented in           */
/*                                                                   */
/*   "A fast exact GLA based on code vector activity detection"      */
/*   IEEE Trans. on Image Processing, 9 (8), 1337-1342, August 2000. */
/*   Timo Kaukoranta, Pasi Frnti and Olli Nevalainen                */
/*                                                                   */
/* Naming conventions used in the code                               */
/*                                                                   */
/*    TS        training set (data objects)                          */
/*    CB        codebook (cluster representatives, centroids)        */
/*    P         partitioning (pointing from TS to CB)                */
/*                                                                   */
/*    p-prefix  pointer, e.g. pTS is pointer to the training set TS  */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "GKM"
#define VersionNumber  "Version 0.11"
#define LastUpdated    "08.09.04"

/* converts error values to MSE values */
#define CALC_MSE(val) (double) (val) / (TotalFreq(pTS) * VectorSize(pTS))

#define WRITE

/*-------------------------------------------------------------------*/

#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>
#include "cb.h"
#include "interfc.h"


/*=================== PROTOTYPES ====================================*/

void CreateInitialSolution(PARTITIONING *pP, CODEBOOK *pCB,
    TRAININGSET *pTS);
llong KMeans(PARTITIONING *pP, CODEBOOK *pCB, TRAININGSET *pTS, 
    int iter);
void OptimalPartition(CODEBOOK *pCB, TRAININGSET *pTS, PARTITIONING *pP,
    int *active, int activeCount, llong *distance);
void OptimalRepresentatives(PARTITIONING *pP, TRAININGSET *pTS,
    CODEBOOK *pCB, int *active, int *activeCount);
int BinarySearch(int *arr, int size, int key);

double SetClock(double* start);
double GetClock(double start);
void PrintHeader(int quietLevel);
void PrintIter(int quietLevel, int iter, double error, double time);
void PrintFooter(int quietLevel, double error, double time);
void GKMInfo(char *info);

/* =================== FUNCTIONS ==================================== */

int PerformGKM(PARTITIONING *pP, CODEBOOK *pCB, CODEBOOK* pLocations,
TRAININGSET *pTS, int kmIter, int quietLevel, int saveIntermediateResults)
{
  CODEBOOK CBprev, CBtmp, CBmin;
  PARTITIONING Ptmp, Pmin;
  int k, i;
  llong error, minError = MAXLLONG;
  double c;
  char str1[64] = {'\0'};
  char str2[64] = {'\0'};

  if ((BookSize(pTS) < BookSize(pCB)) || (kmIter < 0))
  {
    return 1;  // clustering failed
  }

  SetClock(&c);
  PrintHeader(quietLevel);

  /* initial solution */
  CreateInitialSolution(&Pmin, &CBmin, pTS);

#ifdef WRITE
    if (saveIntermediateResults)
    {
      WriteCodebook("1.cb", &CBmin, 1);
      WritePartitioning("1.pa", &Pmin, pTS, 1);
    }
#endif

  /* global K-means iterations; we have previously solved k-clustering 
     problem and we are now solving k+1-clustering problem */
  for (k = 1; k < BookSize(pCB); k++)
  {
    CreateNewCodebook(&CBprev, k, pTS);
    CopyCodebook(&CBmin, &CBprev);  

    FreeCodebook(&CBmin);
    FreePartitioning(&Pmin);
    
    ChangeCodebookSize(&CBprev, k+1);

    CreateNewCodebook(&CBmin, k+1, pTS);
    CreateNewPartitioning(&Pmin, pTS, k+1);
    CreateNewCodebook(&CBtmp, k+1, pTS);
    CreateNewPartitioning(&Ptmp, pTS, k+1);
    
    /* try every data vector (as an k+1th code vector) 
       and pick best one */
    minError = MAXLLONG;
    for (i = 0; i < BookSize(pLocations); i++)
    {
      /* try Vector(pTS,i) */
      CopyCodebook(&CBprev, &CBtmp);
      CopyVector(Vector(pLocations, i), Vector(&CBtmp, k), VectorSize(pTS));
      GenerateOptimalPartitioningGeneral(pTS, &CBtmp, &Ptmp, MSE);

      error = KMeans(&Ptmp, &CBtmp, pTS, kmIter);

      if (error < minError)
      {
        minError = error;
        CopyCodebook(&CBtmp, &CBmin);
        CopyPartitioning(&Ptmp, &Pmin);
      }
    }

    FreeCodebook(&CBprev);
    FreeCodebook(&CBtmp);
    FreePartitioning(&Ptmp);

#ifdef WRITE
    if (saveIntermediateResults)
    {
      sprintf(str1,"%d.cb", k+1);    
      sprintf(str2,"%d.pa", k+1);    
      WriteCodebook(str1, &CBmin, 1);
      WritePartitioning(str2, &Pmin, pTS, 1);
    }
#endif

    PrintIter(quietLevel, k, CALC_MSE(minError), GetClock(c));
  }

  CopyCodebook(&CBmin, pCB);
  CopyPartitioning(&Pmin, pP);

  FreeCodebook(&CBmin);
  FreePartitioning(&Pmin);

  PrintFooter(quietLevel, CALC_MSE(minError), GetClock(c));

  return 0;
}  /* PerformGKM() */

/* ------------------------------------------------------------------ */

/* finds centroid of the data set pTS */
void CreateInitialSolution(PARTITIONING *pP, CODEBOOK *pCB, 
TRAININGSET *pTS)
{
  int i;
  CreateNewCodebook(pCB, 1, pTS);
  CreateNewPartitioning(pP, pTS, 1);
  for (i = 0; i < BookSize(pTS); i++)
  {
    ChangePartition(pTS, pP, 0, i);
  }
  PartitionCentroid(pP, 0, &Node(pCB, 0));
}  /* CreateInitialSolution() */

/*-------------------------------------------------------------------*/

/* fast K-means implementation (uses activity detection method) */
llong KMeans(PARTITIONING *pP, CODEBOOK *pCB, TRAININGSET *pTS, int iter)
{
  int    i, j, activeCount;
  int    active[BookSize(pCB)];
  llong  distance[BookSize(pTS)];
  llong  oldError, error;

  oldError = MAXLLONG;
  error = 0;
  
  /* calculate data objects current distances to their 
     cluster centroids and current error value */
  for (i = 0; i < BookSize(pTS); i++)
  {
    j           = Map(pP, i);
    distance[i] = VectorDistance(Vector(pTS, i), Vector(pCB, j),
                  VectorSize(pTS), MAXLLONG, EUCLIDEANSQ);
    error      += distance[i] * VectorFreq(pTS, i);
  }

  /* performs K-means iterations until no improvement achieved 
     or iter iterations has been made */
  for (i = 0; (i < iter) && (error < oldError); i++) 
  {
    oldError = error;
    
    OptimalRepresentatives(pP, pTS, pCB, active, &activeCount);
    OptimalPartition(pCB, pTS, pP, active, activeCount, distance);
    
    /* calculates current error value */
    error = 0;
    for (j = 0; j < BookSize(pTS); j++)
    {
      error += distance[j] * VectorFreq(pTS, j);
    }    
  }

  return error;
}  /* KMeans() */
                                        
/* ------------------------------------------------------------------ */

/* generates optimal partitioning with respect to a given codebook */
void OptimalPartition(CODEBOOK *pCB, TRAININGSET *pTS, PARTITIONING *pP,
int *active, int activeCount, llong *distance)
{
  CODEBOOK CBact;
  int i, j, k;
  int nearest;
  llong error, dist;

  if (activeCount < 1)
  {
    /* all vectors are static; there is nothing to do */
    return;
  }

  /* creating subcodebook (active clusters) */
  CreateNewCodebook(&CBact, activeCount, pTS);

  for (i = 0; i < activeCount; i++)
  {
    CopyVector(Vector(pCB, active[i]), Vector(&CBact, i), VectorSize(pCB));
  }

  for(i = 0; i < BookSize(pTS); i++)
  {
     j     = Map(pP, i);
     k     = BinarySearch(active, activeCount, j);
     dist  = VectorDistance(Vector(pTS, i), Vector(pCB, j), VectorSize(pTS),
             MAXLLONG, EUCLIDEANSQ);

     if (k < 0)  /* static vector */
     {
       // search subcodebook
       nearest = FindNearestVector(&Node(pTS,i), &CBact, &error, 0,
                 EUCLIDEANSQ);
       nearest = (error < dist) ? active[nearest] : j;
     }
     else if (dist < distance[i])  /* active vector, centroid moved closer */
     {
       // search subcodebook
       nearest = FindNearestVector(&Node(pTS,i), &CBact, &error, k,
                 EUCLIDEANSQ);
       nearest = active[nearest];
     }
     else  /* active vector, centroid moved farther */
     {
       // search full codebook
       nearest = FindNearestVector(&Node(pTS,i), pCB, &error, j,
                 EUCLIDEANSQ);
     }

     if (nearest != j)
     {
       /* closer cluster was found */
       ChangePartition(pTS, pP, nearest, i);
       distance[i] = error;
       /*distance[i] = VectorDistance(Vector(pTS, i), Vector(pCB, nearest),
                     VectorSize(pTS), MAXLLONG, EUCLIDEANSQ);*/
     } 
     else 
     {
       distance[i] = dist;
     }
  }

  FreeCodebook(&CBact);
}  /* OptimalPartition() */

/* ------------------------------------------------------------------ */

/* generates optimal codebook with respect to a given partitioning */
void OptimalRepresentatives(PARTITIONING *pP, TRAININGSET *pTS,
CODEBOOK *pCB, int *active, int *activeCount)
{
  int i, j;
  VECTORTYPE v;

  j = 0;
  v = CreateEmptyVector(VectorSize(pCB));

  for(i = 0; i < BookSize(pCB); i++)
  {
    if (CCFreq(pP, i) > 0)
    {
      CopyVector(Vector(pCB, i), v, VectorSize(pCB));

      /* calculate mean values for centroid */
      PartitionCentroid(pP, i, &Node(pCB, i));

      /* if centroid changed, cluster is active */
      if (CompareVectors(Vector(pCB, i), v, VectorSize(pCB)) != 0)
      {
        active[j] = i;
        j++;
      }
    }
    else
    {
      VectorFreq(pCB, i) = 0;
    }
  }

  FreeVector(v);
  (*activeCount) = j;
}  /* OptimalRepresentatives() */

/* ------------------------------------------------------------------ */

/* arr must be sorted ascending order! */
int BinarySearch(int *arr, int size, int key)
{
  int top, bottom, middle;

  top = 0;
  bottom = size - 1;
  middle = (top + bottom) / 2;

  do
  {
    if (arr[middle] < key)   
    {
      top = middle + 1;
    }
    else
    {
      bottom = middle;
    }

    middle = (top + bottom) / 2;
  }
  while (top < bottom);

  if (arr[middle] == key)
  {
    return middle;
  }
  else
  {
    return -1;
  }
}  /* BinarySearch() */

/* ------------------------------------------------------------------ */

double SetClock(double* start)
{
  struct timeval tv;
  gettimeofday(&tv, NULL);

  return( *start = (double)tv.tv_sec + ((double)tv.tv_usec /1e6) );
}  /* SetClock() */

/* ------------------------------------------------------------------ */

double GetClock(double start)
{
  struct timeval tv;
  double elapsed;

  gettimeofday(&tv, NULL);
  elapsed = (double)tv.tv_sec + ((double)tv.tv_usec /1e6) - start;

  return elapsed;
}  /* GetClock() */

/* ------------------------------------------------------------------ */

void PrintHeader(int quietLevel)
{
  if (quietLevel > 1)
  {
    PrintMessage("\nBook size\tMSE\t\tTime\n");
  }
}  /* PrintHeader() */

/*-------------------------------------------------------------------*/

void PrintIter(int quietLevel, int iter, double error, double time)
{
  if (quietLevel > 1)
  {
    PrintMessage("%d\t\t%f\t%f\n", iter+1, error, time); 
  }  
}  /* PrintIter() */

/*-------------------------------------------------------------------*/
              
void PrintFooter(int quietLevel, double error, double time)
{
  if (quietLevel > 0)
  {
    PrintMessage("\nmse = %f  time = %f\n", error, time);
  }
}  /* PrintHeader() */

/*-------------------------------------------------------------------*/
              
void GKMInfo(char *info)
{
  sprintf(info, "%s   %s   %s", ProgName, VersionNumber, LastUpdated);
}  /* GKMInfo() */

/*-------------------------------------------------------------------*/

